/*
** crypto_api.c
**
** Made by mycroft (mycroft)
** <mycroft@minithins.net> on virgaria.minithins.net
**
** Started on  Sat Jan 25 15:25:59 2003 mycroft
** Last update Sat Jan 25 15:25:59 2003 mycroft
*/

#include <stdio.h>
#include <stdlib.h>
#include "crypto_api.h"
#include "my_functions.h"

static int method = 0;
static char **(*cipher_name)(void);
static u4byte *(*set_key)(const u4byte in_key[], const u4byte key_len); 
static void (*encrypt)(const u4byte in_blk[4], u4byte out_blk[4]);
static void (*decrypt)(const u4byte in_blk[4], u4byte out_blk[4]);

char **clear_cipher_name(void)
{
  static char *alg_name[] = { "clean", "clean.c" };
  return alg_name;
}

u4byte *clear_set_key(const u4byte in_key[], const u4byte key_len)
{
  /* Oui, non rien */
  return (NULL);
}

void clear_encrypt(const u4byte in_blk[4], u4byte out_blk[4])
{
  int i;
  char *t;

  my_memcpy (out_blk, in_blk, sizeof (u4byte) * 4);
  t = (char*)out_blk;
  for (i = 0 ; i < 16; i ++, t++)
    (*t) ++;

  return ;
}

void clear_decrypt(const u4byte in_blk[4], u4byte out_blk[4])
{
  int i;
  char *t;

  my_memcpy (out_blk, in_blk, sizeof (u4byte) * 4);
  t = (char*)out_blk;
  for (i = 0 ; i < 16; i ++, t++)
    (*t) --; 

  return ;
}

int set_method ( int met )
{
  switch ( met )
  { 
#ifdef METHOD_CLEAR
    case METHOD_CLEAR:
      method = METHOD_CLEAR;
      cipher_name = clear_cipher_name;
      set_key = clear_set_key;
      encrypt = clear_encrypt;
      decrypt = clear_decrypt;
    break;
#endif
    case METHOD_CAST:	/* OK */
      method = METHOD_CAST;
      cipher_name = cast_cipher_name;
      set_key = cast_set_key;
      encrypt = cast_encrypt;
      decrypt = cast_decrypt;
    break;
    case METHOD_CRYPTON:	/* OK */
      method = METHOD_CRYPTON;
      cipher_name = crypton_cipher_name;
      set_key = crypton_set_key;
      encrypt = crypton_encrypt;
      decrypt = crypton_decrypt;
    break;
    case METHOD_DEAL:	/* OK */
      method = METHOD_DEAL;
      cipher_name = deal_cipher_name;
      set_key = deal_set_key;
      encrypt = deal_encrypt;
      decrypt = deal_decrypt;
    break;
    case METHOD_DFC:	/* OK */
      method = METHOD_DFC;
      cipher_name = dfc_cipher_name;
      set_key = dfc_set_key;
      encrypt = dfc_encrypt;
      decrypt = dfc_decrypt;
    break;
    case METHOD_E2:	/* OK */
      method = METHOD_E2;
      cipher_name = e2_cipher_name;
      set_key = e2_set_key;
      encrypt = e2_encrypt;
      decrypt = e2_decrypt;
    break;
    case METHOD_HPC:	/* OK */
      method = METHOD_HPC;
      cipher_name = hpc_cipher_name;
      set_key = hpc_set_key;
      encrypt = hpc_encrypt;
      decrypt = hpc_decrypt;
    break;
    case METHOD_MARS:	/* OK */
      method = METHOD_MARS;
      cipher_name = mars_cipher_name;
      set_key = mars_set_key;
      encrypt = mars_encrypt;
      decrypt = mars_decrypt;
    break;
#ifdef METHOD_RC6
    case METHOD_RC6:	/* NOT OK !!! */
      method = METHOD_RC6;
      cipher_name = rc6_cipher_name; 
      set_key = rc6_set_key;
      encrypt = rc6_encrypt;
      decrypt = rc6_decrypt;
    break;
#endif
#ifdef METHOD_RIJNDAEL
    case METHOD_RIJNDAEL:	/* NOT OK !!! */
      method = METHOD_RIJNDAEL;
      cipher_name = rijndael_cipher_name;
      set_key = rijndael_set_key;
      encrypt = rijndael_encrypt;
      decrypt = rijndael_decrypt;
    break;
#endif
    case METHOD_SAFER:	/* OK */
      method = METHOD_SAFER;
      cipher_name = safer_cipher_name;
      set_key = safer_set_key;
      encrypt = safer_encrypt;
      decrypt = safer_decrypt;
    break;
    case METHOD_SERPENT:	/* OK */
      method = METHOD_SERPENT;
      cipher_name = serpent_cipher_name;
      set_key = serpent_set_key;
      encrypt = serpent_encrypt;
      decrypt = serpent_decrypt;
    break;
    case METHOD_TWOFISH: 	/* OK */	
      method = METHOD_TWOFISH;
      cipher_name = twofish_cipher_name;
      set_key = twofish_set_key;
      encrypt = twofish_encrypt;
      decrypt = twofish_decrypt;
    break;
    default: method = -1 ; break;
  }

  return (method);
}

int get_method ( void )
{
  return (method);
}

char **get_cipher_name(void)
{
  return (cipher_name());
}

/* Ok, y a les routines de crypto maintenant */

/* Return value: status
*/

int clear_to_crypted (unsigned char *from, unsigned int from_len,
                      unsigned char **to,  unsigned int *to_len)
{
  int i;
  int from_len_padding;
  unsigned char *from_padding;
  u4byte *tmp1;
  u4byte *tmp2;

  from_len_padding = from_len + (16 - ((from_len % 16)?from_len % 16:16)); 
  from_padding = my_malloc (sizeof (unsigned char) * from_len_padding);
  *to_len = from_len_padding;
  *to = my_malloc (sizeof (unsigned char) * from_len_padding);
  my_memset (*to, 0, from_len_padding);
  my_memcpy (from_padding, from, from_len);

  for (i = 0; i < from_len_padding; i += 16)
  {
    tmp1 = (u4byte *)(from_padding + i);
    tmp2 = (u4byte *)(*to + i);
    encrypt(tmp1, tmp2);
  }

  my_free (from_padding);
  return (0);
}

int crypted_to_clear (unsigned char *from, unsigned int from_len,
                      unsigned char **to,  unsigned int *to_len)
{
  int i;
  int from_len_padding;
  unsigned char *from_padding;
  u4byte *tmp1;
  u4byte *tmp2;

  from_len_padding = from_len + (16 - ((from_len % 16)?from_len % 16:16)); 
  if (from_len_padding != from_len)
    return (-1);
  *to_len = from_len_padding;

  from_padding = my_malloc (sizeof (unsigned char) * from_len_padding);
  *to = my_malloc (sizeof (unsigned char) * from_len_padding);
  my_memset (*to, 0, from_len_padding);
  my_memcpy (from_padding, from, from_len);

  for (i = 0; i < from_len_padding; i += 16)
  {
    tmp1 = (u4byte *)(from_padding + i);
    tmp2 = (u4byte *)(*to + i);
    decrypt(tmp1, tmp2);
  }

  my_free (from_padding);
  return (0);
}


int set_crypt_key (unsigned char *key, int bits)
{
  u4byte *lala = (u4byte*)key;
  set_key (lala, bits);
  return (0);
}
