#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char   u1byte;
typedef unsigned short  u2byte;
typedef unsigned long   u4byte;

typedef signed char     s1byte;
typedef signed short    s2byte;
typedef signed long     s4byte;


#define ENC(Char) (uu_base64[(Char) & 077])

const char uu_base64[64] =
{
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
  'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
  'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
  'w', 'x', 'y', 'z', '0', '1', '2', '3',
  '4', '5', '6', '7', '8', '9', '+', '/'
};

static const char b64_tab[256] =
{
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*000-007*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*010-017*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*020-027*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*030-037*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*040-047*/
  '\177', '\177', '\177', '\76',  '\177', '\177', '\177', '\77',  /*050-057*/
  '\64',  '\65',  '\66',  '\67',  '\70',  '\71',  '\72',  '\73',  /*060-067*/
  '\74',  '\75',  '\177', '\177', '\177', '\100', '\177', '\177', /*070-077*/
  '\177', '\0',   '\1',   '\2',   '\3',   '\4',   '\5',   '\6',   /*100-107*/
  '\7',   '\10',  '\11',  '\12',  '\13',  '\14',  '\15',  '\16',  /*110-117*/
  '\17',  '\20',  '\21',  '\22',  '\23',  '\24',  '\25',  '\26',  /*120-127*/
  '\27',  '\30',  '\31',  '\177', '\177', '\177', '\177', '\177', /*130-137*/
  '\177', '\32',  '\33',  '\34',  '\35',  '\36',  '\37',  '\40',  /*140-147*/
  '\41',  '\42',  '\43',  '\44',  '\45',  '\46',  '\47',  '\50',  /*150-157*/
  '\51',  '\52',  '\53',  '\54',  '\55',  '\56',  '\57',  '\60',  /*160-167*/
  '\61',  '\62',  '\63',  '\177', '\177', '\177', '\177', '\177', /*170-177*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*200-207*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*210-217*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*220-227*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*230-237*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*240-247*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*250-257*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*260-267*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*270-277*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*300-307*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*310-317*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*320-327*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*330-337*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*340-347*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*350-357*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*360-367*/
  '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*370-377*/
};

u1byte *uu_encode (u1byte *string, int len)
{   
  u1byte *p;
  u2byte ch, n;
  u1byte *r_str;
  u1byte c1, c2;

  r_str = (u1byte *) malloc(len * 4 * sizeof (u1byte));
  memset (r_str, 0, len * 4 * sizeof(u1byte));

  n = len;
  /* printf("len is %i\n", n); */
    
  for (p = string; n > 2; n -= 3, p += 3)
  {
    ch = *p >> 2;
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;

    ch = ((*p << 4) & 060) | ((p[1] >> 4) & 017);
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;

    ch = ((p[1] << 2) & 074) | ((p[2] >> 6) & 03);
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;

    ch = p[2] & 077;
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;
  }

  if (n)
  {
    c1 = *p;
    c2 = n == 1 ? 0 : p[1];

    ch = c1 >> 2;
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;

    ch = ((c1 << 4) & 060) | ((c2 >> 4) & 017);
    ch = ENC (ch);
    r_str [strlen((char*)r_str)] = ch;

    if (n == 1)
      ch = '=';
    else
    {
      ch = (c2 << 2) & 074;
      ch = ENC (ch);
      r_str [strlen((char*)r_str)] = ch;
      ch = '=';
    }
    r_str [strlen((char*)r_str)] = ch;
  }
  return (r_str);
}

u1byte *uu_decode (u1byte *string, int *len)
{
  u1byte *r_str, *p, *savp;

  r_str = (u1byte *) malloc (sizeof(u1byte) * strlen((char*)string) + 1);
  p = (u1byte *) malloc (sizeof(u1byte) * strlen((char*)string) + 1);

  savp = p;

  memset (p, 0, strlen((char*)string) + 1);
  memset (r_str, 0, strlen((char*)string) +1);

  *len = 0;

  memcpy (p, string, strlen((char*)string));

  while (*p != 0)
  {
    char c1, c2, c3;

    c1 = b64_tab [*p];
    p ++;
    c2 = b64_tab [*p];
    p ++;

    if (*p == '=')
    {
      r_str[(*len)++] = c1 << 2 | c2 >> 4;
      free (savp); return (r_str);
    }
    c3 = b64_tab[*p];
    p ++;
    r_str[(*len)++] = c1 << 2 | c2 >> 4;
    r_str[(*len)++] = c2 << 4 | c3 >> 2;

    if (*p == '=')
    {
      free (savp); return (r_str);
    }
    else
    {
      r_str[(*len)++] = c3 << 6 | b64_tab[*p];
      p ++;
    }
  }
  free (savp);
  return (r_str);
}
