/*-
 * Copyright (c) 2003, Patrick MARIE <mycroft@virgaria.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: jakk.c,v 1.5 2003/04/22 20:22:12 mycroft Exp $
 * $FreeBSD$
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/types.h>
#include <db.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <sys/termios.h>

#include <readline/readline.h>
#include <readline/history.h>

#include <time.h>

#include <crypto_api.h>
#include <my_functions.h>


#define DEF_PROMPT	"jakk# "
#ifndef _PATH_DATABASE
#warning "Please define _PATH_DATABASE. make _PATH_DATABASE="\\\"file_path\\\""
#define _PATH_DATABASE "/home/mycroft/private/key.db"
#endif

static char *crypt_key;
static DB *database;
recno_t no;

typedef struct t_rec {
	char name[32];
	char pass[32];
	char desc[512];
	unsigned int crc;
	char misc[12];
} __rec; 

char *cmds[] = {
	"add",
	"del",
	"edit",
	"exit",
	"get",
	"help",
	"list",
	"search",
	"usekey",
	NULL
};

char *
get_next_word(char **str)
{
	char *ret, *n;
	if (!*str)
		return (NULL);

	/* remove space */
	while (**str == ' ') 
		(*str) ++;

	/* ok, lets copy the string */
	if (!(n = strchr (*str, ' ')))
		return (strdup (*str));
	else {
		*n = 0;
		ret = strdup (*str);
		*str = n + 1;
		return (ret);
	}

	return (NULL);
}

char *
get_pw(void)
{
	int l;
	struct termios t;
#define BUF_SIZE	512
	char *n = malloc (BUF_SIZE * sizeof (char));

	if (ioctl (0, TIOCGETA, &t) < 0)
		goto bad;

	t.c_lflag &= ~ECHO;
	if (ioctl(0, TIOCSETA, &t) < 0)
		goto bad;

	fprintf(stdout, "password: ");
	fflush(stdout);
	memset (n, 0, BUF_SIZE);
	if ((l = read(0, n, BUF_SIZE - 1)) <= 0) 
		goto bad;
	n[l - 1] = 0;
	write (1, "\n", 1);

	t.c_lflag |= ECHO;

	if (ioctl(0, TIOCSETA, &t) < 0)
		goto bad;

	goto ge;

bad:
	memset (n, 0, BUF_SIZE);
	free(n);
	n = NULL;

ge:
	return n;
}

int
set_crypto_key(void)
{
	int oct;
/*	char *verif_key = NULL; */

	if (crypt_key != NULL) {
		memset (crypt_key, 0, BUF_SIZE);
		free (crypt_key);
	} 

/*
	do
	{
		if (verif_key != NULL)
			free (verif_key);
		if (crypt_key != NULL)
			free (crypt_key);

		fprintf(stdout, "Please input your private key.\n");
		crypt_key = get_pw ();
		fprintf(stdout, "Verification.\n");
		verif_key = get_pw ();

		if (crypt_key == NULL || verif_key == NULL) {
			if (verif_key != NULL)
				free (verif_key);
			if (crypt_key != NULL)
				free (crypt_key);
			return -1;
		}

	} while ((strcmp(crypt_key,verif_key)) != 0);

	if (verif_key != NULL)
		free (verif_key);
*/

	fprintf(stdout, "Please input your private key.\n");
	crypt_key = get_pw ();

	if (crypt_key != NULL) {
		oct = strlen (crypt_key);
		oct *= 8;
		if (oct < 128) {
			oct = 0;
			free (crypt_key);
			crypt_key = NULL;
			return -1;
		} else if (oct < 192) {
			oct = 128;
		} else if (oct < 256) {
			oct = 192;
		} else 
			oct = 256;

		printf ("Setting %d bits key ...\n", oct);
		set_crypt_key (crypt_key, oct);
	}

	return 0;
}

int
record_list(void)
{
	DBT data;
	DBT key;
	int ret;
	unsigned char *u, *ud;
	int lu, niah;
	__rec z;

	if (database == NULL)
		return -1;

	ret = database->seq (database, &key, &data, R_FIRST);
	if (ret != 0) {
		return -1;
	}

	do
	{
		ud = uu_decode(data.data, &lu);
		crypted_to_clear (ud, lu, &u, &niah);
		memcpy (&z, u, sizeof (struct t_rec));

		if (ud != NULL)
			free (ud);
		if (u != NULL)
			free (u);
		
		if (z.crc != crc32 (z.name, sizeof (z.name))) {
			printf ("Invalid CRC for this record, skipping.\n");
			continue ;
		}

		printf ("%2d/  %s\n", *(int*)key.data, z.name);
	}
	while ((ret = database->seq (database, &key, &data, R_NEXT)) == 0);

	return 0;
}

int
record_search(char *str)
{
	DBT data, key;
	int ret, lu, niah;
	__rec z;
	unsigned char *u, *ud;

	ret = database->seq (database, &key, &data, R_FIRST);
	if (ret != 0) {
		return -1;
	}

	do
	{
		ud = uu_decode(data.data, &lu);
		crypted_to_clear (ud, lu, &u, &niah);
		memcpy (&z, u, sizeof (struct t_rec));
		if (ud != NULL)
			free (ud);
		if (u != NULL)
			free (u);

		if (z.crc != crc32 (z.name, sizeof (z.name))) {
			continue ;
		}

		if (strcasestr (z.name, str) != NULL
		 || strcasestr (z.desc, str) != NULL) {
			printf ("%2d/  %s\n", *(int*)key.data, z.name);
			if (strlen (z.desc) > 0)
				printf ("%s\n", z.desc);
			printf ("Key: [%s]\n", z.pass);
		}
	}
	while ((ret = database->seq (database, &key, &data, R_NEXT)) == 0);

	return 0;
}

int
record_get(char *str)
{
	DBT key, data;
	int ret, lu, niah;
	unsigned char *u, *ud;
	__rec z ;

	int l = (int)strtol(str, (char **)NULL, 10);
	if (l == 0)
		return -2;

	no = l;
	key.data = (char *)&no;
	key.size = sizeof (recno_t);

	ret = database->seq (database, &key, &data, R_CURSOR);

	if (ret != 0)
		return -2;

	ud = uu_decode(data.data, &lu);
	crypted_to_clear (ud, lu, &u, &niah);
	memcpy (&z, u, sizeof (struct t_rec));

	if (z.crc != crc32 (z.name, sizeof (z.name))) {
		printf ("Invalid CRC for this record, skipping.\n");
	} else {
		printf ("%2d/  %s\n", *(int*)key.data, z.name);
		if (strlen (z.desc) > 0)
			printf ("%s\n", z.desc);
		printf ("Key: [%s]\n", z.pass);
	}

	return 0;
}

char *
copy_and_fill(char *dst, const char *src, size_t l)
{
	/* Ok, we feel dst with l random things */
	unsigned int i;
	size_t len_cp;

	srandom(time(NULL));

	for (i = 0; i < l; i += 4) {
		*(long*)(dst + i) = random ();
	}

	len_cp = strlen (src);

	strncpy(dst, src, (l > len_cp + 1)?len_cp+1:l);

	if (l <= len_cp + 1)
		memset (dst + l - 1, 0, 1); 

	return (dst);
}

int
record_add(void)
{
	__rec ll;
	unsigned char *t, *crp, *tmp;
	int lcrp;
	DBT ndata, key;
	int ret;

	if (database == NULL)
		return -1;

	fprintf (stdout, "IndexName: "); fflush (stdout);
	t = get_next_line (0);
	copy_and_fill (ll.name, t, sizeof (ll.name));
	free (t);

	fprintf (stdout, "Description: "); fflush (stdout);
	t = get_next_line (0);
	copy_and_fill (ll.desc, t, sizeof (ll.desc));
	free (t);

	fprintf (stdout, "PassWord: "); fflush (stdout);
	t = get_next_line (0);
	copy_and_fill (ll.pass, t, sizeof (ll.pass));
	free (t);

	ll.crc = crc32 (ll.name, sizeof (ll.name)); 

	ret = clear_to_crypted ((char*)&ll, sizeof (struct t_rec), &crp, &lcrp);
	if (ret == -1)
		printf ("NO DATA ENCODING !\n");
	tmp = uu_encode(crp, lcrp);

    ret = database->seq (database, &key, &ndata, R_LAST);
    if (ret != 0) {
		no = 0;
    }

	ndata.data = (void*)tmp;
	ndata.size = strlen (tmp) + 1; 

	no ++;
	key.data = (char *)&no;
	key.size = sizeof (recno_t);

	ret = database->put (database, &key, &ndata, R_NOOVERWRITE);

	return 0;
}

int
record_edit(char *str)
{
	DBT key, data;
	DBT ndata;
	int ret, lu, niah;
	unsigned char *u, *ud, *t;
	__rec z ;
	unsigned char *crp, *tmp;
	int lcrp;

	int l = (int)strtol(str, (char **)NULL, 10);
	if (l == 0)
		return -2;

	no = l;
	key.data = (char *)&no;
	key.size = sizeof (recno_t);

	ret = database->seq (database, &key, &data, R_CURSOR);

	if (ret != 0)
		return -2;

	ud = uu_decode(data.data, &lu);
	crypted_to_clear (ud, lu, &u, &niah);
	memcpy (&z, u, sizeof (struct t_rec));

	if (z.crc != crc32 (z.name, sizeof (z.name))) {
		printf ("Invalid CRC for this record, skipping.\n");
		return -1;
	} else {
		/* Data ici */
		fprintf (stdout, "IndexName [%s]\n", z.name);
		fprintf (stdout, "IndexName: "); fflush (stdout);
		if ((t = get_next_line (0))) {
			if (strlen(t) != 0)
				copy_and_fill (z.name, t, sizeof (z.name));
			free(t);
		}

		fprintf (stdout, "Description: [%s]\n", z.desc);
		fprintf (stdout, "Description: "); fflush (stdout);
		if ((t = get_next_line (0))) {
			if (strlen(t) != 0)
				copy_and_fill (z.desc, t, sizeof (z.desc));
			free (t);
		}

		fprintf (stdout, "PassWord: [%s]\n", z.pass);
		fprintf (stdout, "PassWord: "); fflush (stdout);
		if ((t = get_next_line (0))) {
			if (strlen(t) != 0)
				copy_and_fill (z.pass, t, sizeof (z.pass));
			free(t);
		}

		z.crc = crc32 (z.name, sizeof (z.name));

		ret = clear_to_crypted ((char*)&z, sizeof (struct t_rec), &crp, &lcrp);
		if (ret == -1)
			printf ("NO DATA ENCODING !\n");
		tmp = uu_encode(crp, lcrp);

		ndata.data = (void*)tmp;
		ndata.size = strlen (tmp) + 1;

		/* Sauvegarde */
		ret = database->put (database, &key, &ndata, R_CURSOR);
	}

	return 0;
}

int
record_del(char *str)
{
	DBT key, data;
	int ret;

	int l = (int)strtol(str, (char **)NULL, 10);
	if (l == 0)
		return -2;

	no = l;
	key.data = (char *)&no;
	key.size = sizeof (recno_t);

	ret = database->seq (database, &key, &data, R_CURSOR);
	if (ret != 0) {
		fprintf (stderr, "Unable to find this record.\n");
		return -2;
	}

	ret = database->del (database, &key, R_CURSOR);
	if (ret != 0) {
		fprintf (stderr, "Unable to delete this record.\n");
		return -2;
	}

	fprintf (stdout, "Record deleted.\n");
	return 0;
}

int
act_line(char *str)
{
	char *cmd = get_next_word(&str);
	char *arg1 = get_next_word(&str);

	if (!cmd)
		return -1;

	if (!strcmp (cmd, "help")) {
		fprintf (stdout, "Help of jakk\n\n"
		"usekey    -  define new key\n"
		"list      -  list all index keys\n"
		"get X     -  get record number X\n"
		"del X     -  del record number X\n"
		"search X  -  search record with patter X\n"
		"add       -  add record\n"
		"exit      -  exit jakk\n\n");
	} else if (!strcmp (cmd, "usekey")) {
		return set_crypto_key ();
	} else if (!strcmp (cmd, "list")) {
		return record_list (); /* DONE */
	} else if (!strcmp (cmd, "add")) {
		return record_add (); /* DONE */
	} else if (!strcmp (cmd, "get")) {
		return record_get (arg1); /* DONE */
	} else if (!strcmp (cmd, "del")) {
		record_del (arg1); /* DONE */ 
	} else if (!strcmp (cmd, "search")) {
		record_search (arg1); /* DONE */
	} else if (!strcmp (cmd, "edit")) {
		record_edit (arg1);
	} else if (!strcmp (cmd, "exit")) {
		return -3;
	}

	return 0;
}

int
open_database(void)
{
	int ret;
	DBT key, data;

	if (database != NULL)
		return -2;

	database = dbopen (_PATH_DATABASE, O_CREAT | O_RDWR, 0600, DB_RECNO, NULL);
	if (database == NULL)
		return -1;

	ret = database->seq (database, &key, &data, R_LAST);
	if (ret != 0) {
		no = 0;
	} else {
		no = *(int*)key.data;
	}
	printf ("last rec num is %d\n", no);

	return 0;
}

int
close_database(void)
{
	if (database != NULL) {
		database->close (database);
		database = NULL;
		return 0;
	}
	return -1;
}

char *command_completion(char *text, int state)
{
	char *name;
	static char **tmp = cmds;

	while (*tmp) {

		if (!strncasecmp(text, *tmp, strlen(text))) {
			name = strdup (*tmp);
			tmp ++;
			return (name);
		}

		tmp ++;
	}

	tmp = cmds;

	return (NULL);
}

char **complete(char *text, int start, int end)
{
	char **matches;

	matches = rl_completion_matches(text,
	  (rl_compentry_func_t*)command_completion);

	return (matches);
}

int
main(int argc, char **argv)
{
	char *string;
	char *prompt = DEF_PROMPT;

	open_database ();
	set_method (METHOD_TWOFISH /* dek is my lover */);
	set_crypto_key ();

	rl_attempted_completion_function = (rl_completion_func_t*)complete;

	while ((string = readline (prompt)) != NULL) {
		add_history (string);
		switch (act_line(string)) {
		case -1: fprintf (stderr, "Command error.\n"); break;
		case -2: fprintf (stderr, "Argument error.\n"); break;
		case -3: goto end; break;
		case -4: fprintf (stderr, "No command found.\n"); break;
		default: /* Nothing */ break;
		}
	}
	printf ("\n");

end:
	close_database ();
	/* slaps the crypt key */
	if (crypt_key)
		memset (crypt_key, 0, BUF_SIZE);
	fprintf(stdout, "\nEnd of jakk.\n");

	return 0;
}
